#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

import inspect
from typing import List, Union

from bs4 import BeautifulSoup
from requests.models import Response

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.evidence import Evidence
from yawast.scanner.plugins.http import http_basic, retirejs, error_checker, http_utils
from yawast.scanner.plugins.http.servers import rails, apache_tomcat
from yawast.scanner.plugins.result import Result
from yawast.shared import network, output


def check_response(
    url: str, res: Response, soup: Union[BeautifulSoup, None] = None
) -> List[Result]:
    # make sure we actually have something
    if res is None:
        return []

    results: List[Result] = []

    raw_full = network.http_build_raw_response(res)

    if http_utils.is_text(res):
        body = res.text

        if soup is None:
            soup = BeautifulSoup(body, "html.parser")

        # check for things thar require parsed HTML
        results += retirejs.get_results(soup, url, res)
        results += apache_tomcat.get_version(url, res)
        results += error_checker.check_response(url, res, body)

        results += _check_cache_headers(url, res)

    results += http_basic.get_header_issues(res, raw_full, url)
    results += http_basic.get_cookie_issues(res, url)

    # this function will trigger a recursive call, as it calls this to check the response.
    # to deal with this, we'll check the caller, to make sure it's not what we're about to call.
    if "check_cve_2019_5418" not in inspect.stack()[1].function:
        results += rails.check_cve_2019_5418(url)

    results += _check_charset(url, res)

    return results


def _check_charset(url: str, res: Response) -> List[Result]:
    results: List[Result] = []

    # if the body is empty, we really don't care about this
    if len(res.content) == 0:
        return results

    try:
        if "Content-Type" in res.headers:
            content_type = str(res.headers["Content-Type"]).lower()

            if "charset" not in content_type and "text/html" in content_type:
                # not charset specified
                results.append(
                    Result.from_evidence(
                        Evidence.from_response(res, {"content-type": content_type}),
                        f"Charset Not Defined in '{res.headers['Content-Type']}' at {url}",
                        Vulnerabilities.HTTP_HEADER_CONTENT_TYPE_NO_CHARSET,
                    )
                )
        else:
            # content-type missing
            results.append(
                Result.from_evidence(
                    Evidence.from_response(res),
                    f"Content-Type Missing: {url} ({res.request.method} - {res.status_code})",
                    Vulnerabilities.HTTP_HEADER_CONTENT_TYPE_MISSING,
                )
            )
    except Exception:
        output.debug_exception()

    return results


def _check_cache_headers(url: str, res: Response) -> List[Result]:
    results = []

    try:
        if "Cache-Control" in res.headers:
            # we have the header, check the content
            if "public" in str(res.headers["Cache-Control"]).lower():
                results.append(
                    Result.from_evidence(
                        Evidence.from_response(res),
                        f"Cache-Control: Public: {url}",
                        Vulnerabilities.HTTP_HEADER_CACHE_CONTROL_PUBLIC,
                    )
                )

            if "no-cache" not in str(res.headers["Cache-Control"]).lower():
                results.append(
                    Result.from_evidence(
                        Evidence.from_response(res),
                        f"Cache-Control: no-cache Not Found: {url}",
                        Vulnerabilities.HTTP_HEADER_CACHE_CONTROL_NO_CACHE_MISSING,
                    )
                )

            if "no-store" not in str(res.headers["Cache-Control"]).lower():
                results.append(
                    Result.from_evidence(
                        Evidence.from_response(res),
                        f"Cache-Control: no-store Not Found: {url}",
                        Vulnerabilities.HTTP_HEADER_CACHE_CONTROL_NO_STORE_MISSING,
                    )
                )

            if "private" not in str(res.headers["Cache-Control"]).lower():
                results.append(
                    Result.from_evidence(
                        Evidence.from_response(res),
                        f"Cache-Control: private Not Found: {url}",
                        Vulnerabilities.HTTP_HEADER_CACHE_CONTROL_PRIVATE_MISSING,
                    )
                )
        else:
            # header missing
            results.append(
                Result.from_evidence(
                    Evidence.from_response(res),
                    f"Cache-Control Header Not Found: {url}",
                    Vulnerabilities.HTTP_HEADER_CACHE_CONTROL_MISSING,
                )
            )

        if "Expires" not in res.headers:
            results.append(
                Result.from_evidence(
                    Evidence.from_response(res),
                    f"Expires Header Not Found: {url}",
                    Vulnerabilities.HTTP_HEADER_EXPIRES_MISSING,
                )
            )

        else:
            # TODO: parse the value and see if it's less than now
            pass

        if "Pragma" not in res.headers or "no-cache" not in str(res.headers["Pragma"]):
            results.append(
                Result.from_evidence(
                    Evidence.from_response(res),
                    f"Pragma: no-cache Not Found: {url}",
                    Vulnerabilities.HTTP_HEADER_PRAGMA_NO_CACHE_MISSING,
                )
            )
    except Exception:
        output.debug_exception()

    return results



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

import inspect
import logging
import shutil
import sys
import textwrap
import traceback
from inspect import FrameInfo
from multiprocessing.util import get_logger
from threading import Lock
from typing import cast, Optional

from colorama import init, Fore, Style

from yawast.reporting import reporter
from yawast.shared import utils

_no_colors = False
_no_wrap = False
_init = False
_wrapper = None
_debug = False
_logger: Optional[logging.Logger] = None
_lock = Lock()


def setup(enable_debug: bool, no_colors: bool, no_wrap: bool):
    global _no_colors, _init, _wrapper, _debug, _logger, _no_wrap

    _init = True

    _no_wrap = no_wrap

    _wrapper = textwrap.TextWrapper()
    width = shutil.get_terminal_size().columns
    _wrapper.width = width if width > 0 else 80
    _wrapper.subsequent_indent = "\t\t\t\N{DOWNWARDS ARROW WITH TIP RIGHTWARDS} "
    _wrapper.tabsize = 4
    _wrapper.drop_whitespace = False

    # setup the root logger
    rt = logging.getLogger()
    rt.addHandler(_LogHandler())
    rt.setLevel(logging.DEBUG)

    # setup our logger
    _logger = logging.getLogger("yawast")
    _logger.setLevel(logging.DEBUG)
    _logger.addHandler(_LogHandler())
    _logger.propagate = False

    # setup the logger for multiprocessing
    lg = get_logger()
    lg.level = logging.DEBUG
    lg.addHandler(_LogHandler())

    if not no_colors:
        init()
    else:
        _no_colors = True

    if enable_debug:
        toggle_debug()


def is_debug() -> bool:
    global _debug

    return _debug


def toggle_debug():
    global _debug, _logger, _lock

    with _lock:
        _debug = not _debug


def empty():
    print("")


def norm(msg: str):
    val = str(msg)

    _print("       " + val.expandtabs(tabsize=3))


def info(msg: str):
    val = str(msg)

    _print_special(Fore.GREEN, "Info", val)


def warn(msg: str):
    val = str(msg)

    _print_special(Fore.YELLOW, "Warn", val)


def vuln(msg: str):
    val = str(msg)

    _print_special(Fore.RED, "Vuln", val)


def error(msg: str):
    val = str(msg)

    _print_special(Fore.MAGENTA, "Error", val)


def debug(msg: str):
    global _init, _logger

    if _init:
        fi = cast(FrameInfo, inspect.stack()[1])
        val = str(f"{fi.function}:{msg}")

        if _logger is not None:
            _logger.debug(val)


def debug_exception():
    global _init

    if _init:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        val = traceback.format_exception(exc_type, exc_value, exc_traceback)

        debug("Exception: " + "".join(val))


def print_color(color: Fore, msg: str):
    global _no_colors

    if _no_colors:
        _print(msg)
    else:
        _print(color + msg + Style.RESET_ALL)


def _internal_debug(msg: str):
    val = str(msg)

    _print_special(Fore.BLUE, "Debug", val)


def _print_special(color: str, header: str, msg: str):
    global _no_colors

    if _no_colors:
        _print("[{header}] {msg}".format(header=header, msg=msg.expandtabs(tabsize=3)))
    else:
        _print(
            color
            + Style.BRIGHT
            + "[{}] ".format(header)
            + Style.RESET_ALL
            + msg.expandtabs(tabsize=3)
        )


def _print(val):
    global _wrapper, _lock, _debug, _no_wrap

    is_dbg = False

    # we wrap this in a lock, to keep the output clean
    with _lock:
        # register the message with the reporter
        clean = utils.strip_ansi_str(val)
        if clean.startswith("[Debug]"):
            reporter.register_message(clean, "debug")

            is_dbg = True
        else:
            reporter.register_message(clean, "normal")

        if not is_dbg or (is_dbg and _debug):
            if _no_wrap:
                print(val)
            else:
                print(_wrapper.fill(val))


class _LogHandler(logging.StreamHandler):
    def __init__(self):
        logging.Handler.__init__(self)

        self.stream = sys.stderr

        self.setFormatter(
            logging.Formatter(
                fmt="{asctime} {name}:{process}:{threadName}:{filename}:{lineno}: {message}",
                style="{",
            )
        )

    def emit(self, record):
        try:
            msg = self.format(record)

            _internal_debug(msg)
        except RecursionError:
            raise
        except Exception:
            self.handleError(record)



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

import socket

from yawast.scanner.cli import dns
from yawast.scanner.session import Session


def start(session: Session):
    print(f"Scanning: {session.url}")

    # make sure it resolves
    try:
        socket.gethostbyname(session.domain)
    except socket.gaierror as error:
        print(f"Fatal Error: Unable to resolve {session.domain} ({str(error)})")

        return

    dns.scan(session)



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

import distutils
import opcode
import sys
from os import path
from pathlib import Path
from typing import List, Union, Tuple

from requirementslib import Lockfile
from setuptools import find_packages

if "build_exe" in sys.argv:
    from cx_Freeze import setup, Executable
else:
    from setuptools import setup

    # fake Executable class to avoid cx_Freeze on non-Windows
    # noinspection Mypy
    class Executable:
        # noinspection PyUnusedLocal
        def __init__(self, script=None, base=None):
            pass


# effectively a NOP, to keep the import
# we are doing this as a hack, to fix a cx_Freeze issue.
_ = distutils.__version__
distutils_path = path.join(path.dirname(opcode.__file__), "distutils")

if getattr(sys, "frozen", False):
    # frozen
    root_path = path.dirname(sys.executable)
else:
    # unfrozen
    root_path = path.dirname(path.realpath(__file__))

# search for DLL files needed for SSL support
if hasattr(sys, "real_prefix"):
    win_search_path = sys.real_prefix
else:
    win_search_path = sys.base_prefix

win_search_pattern = ["**/libcrypto-*.dll", "**/libssl-*.dll"]
win_include_files: List[Union[str, Tuple[Union[bytes, str], str]]] = []
for srch in win_search_pattern:
    for dll in Path(win_search_path).glob(srch):
        full_path = str(dll.resolve(True))
        target_name = path.join("lib", path.basename(full_path))
        win_include_files.append((full_path, target_name))
        break

win_include_files.append((distutils_path, "distutils"))

# Dependencies are automatically detected.
# I'm not sure about the *version.py files, but this hack works.
build_exe_options = {
    "packages": ["os", "dns"],
    "includes": [
        "six",
        "appdirs",
        "packaging.version",
        "packaging.specifiers",
        "packaging.requirements",
        "html.parser",
        "setuptools.msvc",
        "psutil",
        "colorama",
        "validator_collection",
        "jsonschema",
        "cryptography",
        "_cffi_backend",
        "requests",
        "requests.packages.idna",
        "requests.packages.idna.idnadata",
        "idna.idnadata",
        "publicsuffixlist",
        "sslyze",
        "bs4",
        "selenium",
        "configparser",
        "ordered_set",
    ],
    "excludes": ["tkinter", "distutils"],
    "include_files": win_include_files,
}
bdist_msi_options = {"add_to_path": True}


def get_version_and_cmdclass(package_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location("version", os.path.join(package_path, "_version.py"))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


version, cmdclass = get_version_and_cmdclass("yawast")


def get_long_description():
    """Convert the README file into the long description.
    """
    with open(path.join(root_path, "README.md"), encoding="utf-8") as f:
        long_description = f.read()
    return long_description


def get_install_reqs():
    try:
        lf = Lockfile.load(root_path)

        return lf.requirements_list
    except AttributeError:
        # if it fails, return an empty list - lock file likely missing
        return []


setup(
    name="yawast",
    version=version,
    cmdclass=cmdclass,
    description="The YAWAST Antecedent Web Application Security Toolkit",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/adamcaudill/yawast",
    project_urls={
        "Bug Reports": "https://github.com/adamcaudill/yawast/issues",
        "Source": "https://github.com/adamcaudill/yawast",
        "Changelog": "https://github.com/adamcaudill/yawast/blob/master/CHANGELOG.md",
    },
    author="Adam Caudill",
    author_email="adam@adamcaudill.com",
    license="MIT",
    options={"build_exe": build_exe_options, "build_msi": bdist_msi_options},
    executables=[Executable("bin/yawast", base=None)],
    packages=find_packages(exclude=["tests"]),
    scripts=["bin/yawast"],
    install_requires=get_install_reqs(),
    include_package_data=True,
    package_data={"yawast": ["resources/*"]},
    zip_safe=False,
    python_requires=">=3.7",
    keywords="security tls ssl dns http scan vulnerability",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Natural Language :: English",
        "Topic :: Internet :: Name Service (DNS)",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Networking",
        "Topic :: Software Development :: Testing",
        "Topic :: Security",
    ],
)



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

import socket

from yawast.commands import utils as cutils
from yawast.scanner.cli import dns, ssl_labs, ssl_internal, ssl_sweet32, http, network
from yawast.scanner.session import Session
from yawast.shared import utils, output


def start(session: Session):
    print(f"Scanning: {session.url}")

    # make sure it resolves
    try:
        socket.gethostbyname(session.domain)
    except socket.gaierror as error:
        output.debug_exception()
        output.error(f"Fatal Error: Unable to resolve {session.domain} ({str(error)})")

        return

    try:
        cutils.check_redirect(session)
    except Exception as error:
        output.debug_exception()
        output.error(f"Unable to continue: {str(error)}")

        return

    if not session.args.nodns:
        dns.scan(session)

    if session.args.ports:
        network.scan(session)

    # check to see if we are looking at an HTTPS server
    if session.url_parsed.scheme == "https" and not session.args.nossl:
        if (
            session.args.internalssl
            or utils.is_ip(session.domain)
            or utils.get_port(session.url) != 443
        ):
            # use internal scanner
            ssl_internal.scan(session)
        else:
            try:
                ssl_labs.scan(session)
            except Exception as error:
                output.debug_exception()

                output.error(f"Error running scan with SSL Labs: {str(error)}")
                output.norm("Switching to internal SSL scanner...")

                ssl_internal.scan(session)

        if session.args.tdessessioncount:
            ssl_sweet32.scan(session)

    http.scan(session)

    # reset any stored data
    http.reset()

    return



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

from typing import Union, Dict

from packaging import version

from yawast.shared import network, output

_versions: Union[Dict[str, Dict[str, str]], None] = {}


def get_latest_version(
    package: str, base_version: Union[str, version.Version]
) -> Union[version.Version, None]:
    global _versions

    if _versions is not None:
        # make sure that we have data loaded
        if len(_versions) == 0:
            _get_version_data()

        if package in _versions:
            # check the type of base_version, and parse as needed
            if type(base_version) is version.Version:
                base_version = ".".join(str(base_version).split(".")[0:2])

            if base_version in _versions[package]:
                return version.parse(_versions[package][base_version])
            else:
                return version.parse(_versions[package]["latest"])
        else:
            return None
    else:
        # if it's none, that means that we've attempted to get the version data, and it failed
        output.debug(
            f"_versions is None; skipping version check for {package}:{base_version}"
        )

        return None


def _get_version_data() -> None:
    global _versions
    data: Union[Dict[str, Dict[str, Dict[str, str]]], None] = None
    data_url = "https://raw.githubusercontent.com/adamcaudill/current_versions/master/current_versions.json"

    try:
        data, _ = network.http_json(data_url)
    except Exception as error:
        output.debug(f"Failed to get version data: {error}")
        output.debug_exception()

    if data is not None and "software" in data:
        _versions = data["software"]
    else:
        _versions = None



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

import time
from multiprocessing import Manager, Lock
from multiprocessing.dummy import Pool
from typing import List, Tuple
from urllib.parse import urljoin

from bs4 import BeautifulSoup

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.evidence import Evidence
from yawast.scanner.plugins.http import response_scanner, http_utils
from yawast.scanner.plugins.result import Result
from yawast.shared import network, output

_links: List[str] = []
_insecure: List[str] = []
_lock = Lock()
_tasks = []


def spider(url) -> Tuple[List[str], List[Result]]:
    global _links, _insecure, _tasks, _lock

    results: List[Result] = []

    # create processing pool
    pool = Pool()
    mgr = Manager()
    queue = mgr.Queue()

    asy = pool.apply_async(_get_links, (url, [url], queue, pool))

    with _lock:
        _tasks.append(asy)

    while True:
        if all(t is None or t.ready() for t in _tasks):
            break
        else:
            count_none = 0
            count_ready = 0
            count_not_ready = 0

            for t in _tasks:
                if t is None:
                    count_none += 1
                elif t.ready():
                    count_ready += 1
                else:
                    count_not_ready += 1

            output.debug(
                f"Spider Task Status: None: {count_none}, Ready: {count_ready}, Not Ready: {count_not_ready}"
            )

        time.sleep(3)

    pool.close()

    for t in _tasks:
        try:
            t.get()
        except Exception:
            output.debug_exception()

    while not queue.empty():
        res = queue.get()

        if len(res) > 0:
            for re in res:
                if re not in results:
                    results.append(re)

    # copy data and reset
    links = _links[:]
    _links = []
    _insecure = []

    return links, results


def _get_links(base_url: str, urls: List[str], queue, pool):
    global _links, _insecure, _tasks, _lock

    max_length = 1024 * 1024 * 3  # 3MB

    results: List[Result] = []

    # fail-safe to make sure we don't go too crazy
    if len(_links) > 10000:
        # if we have more than 10,000 URLs in our list, just stop
        output.debug(
            "Spider: Link list contains > 10,000 items. Stopped gathering more links."
        )

        return

    for url in urls:
        try:
            # list of pages found that will need to be processed
            to_process: List[str] = []

            res = network.http_get(url, False)

            # get the length, so that we don't parse huge documents
            if "Content-Length" in res.headers:
                length = int(res.headers["Content-Length"])
            else:
                length = len(res.content)

            if http_utils.is_text(res) and length < max_length:
                soup = BeautifulSoup(res.text, "html.parser")
            else:
                # no clue what this is
                soup = None

            results += response_scanner.check_response(url, res, soup)

            if soup is not None:
                for link in soup.find_all("a"):
                    href = link.get("href")

                    if str(href).startswith("/") and not str(href).startswith("//"):
                        href = urljoin(base_url, href)

                    if href is not None:
                        # check to see if this link is in scope
                        if base_url in href and href not in _links:
                            if "." in href.split("/")[-1]:
                                file_ext = href.split("/")[-1].split(".")[-1]
                            else:
                                file_ext = None

                            with _lock:
                                _links.append(href)

                            # filter out some of the obvious binary files
                            if file_ext is None or file_ext not in [
                                "gzip",
                                "jpg",
                                "jpeg",
                                "gif",
                                "woff",
                                "zip",
                                "exe",
                                "gz",
                                "pdf",
                            ]:
                                if not _is_unsafe_link(href, link.string):
                                    to_process.append(href)
                                else:
                                    output.debug(
                                        f"Skipping unsafe URL: {link.string} - {href}"
                                    )
                            else:
                                output.debug(
                                    f'Skipping URL "{href}" due to file extension "{file_ext}"'
                                )
                        else:
                            if (
                                base_url.startswith("https://")
                                and str(href).startswith("http://")
                                and str(href) not in _insecure
                            ):
                                # link from secure to insecure
                                with _lock:
                                    _insecure.append(str(href))

                                results.append(
                                    Result.from_evidence(
                                        Evidence.from_response(res, {"link": href}),
                                        f"Insecure Link: {url} links to {href}",
                                        Vulnerabilities.HTTP_INSECURE_LINK,
                                    )
                                )

            # handle redirects
            if "Location" in res.headers:
                redirect = res.headers["Location"]

                # check for relative link
                if str(redirect).startswith("/"):
                    redirect = urljoin(base_url, redirect)

                # make sure that we aren't redirected out of scope
                if str(redirect).startswith(base_url):
                    to_process.append(redirect)

            if len(to_process) > 0:
                asy = pool.apply_async(_get_links, (base_url, to_process, queue, pool))

                with _lock:
                    _tasks.append(asy)
        except Exception:
            output.debug_exception()

    output.debug(f"GetLinks Task Completed - {len(results)} issues found.")
    queue.put(results)


def _is_unsafe_link(href: str, description: str) -> bool:
    """
    Check for strings that indicate an unsafe link
    :param href:
    :param description:
    :return:
    """
    unsafe_fragments = [
        "logoff",
        "log off",
        "log_off",
        "logout",
        "log out",
        "log_out",
        "delete",
        "destroy",
    ]

    ret = False

    try:
        description = str(description).lower() if description is not None else ""
        href = str(href).lower()

        for frag in unsafe_fragments:
            if frag in href or frag in description:
                return True
    except Exception:
        output.debug_exception()

    return ret



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

from typing import Optional

from yawast.external.spinner import Spinner
from yawast.reporting import reporter
from yawast.scanner.plugins.dns import basic
from yawast.scanner.plugins.network import port_scan
from yawast.scanner.session import Session
from yawast.shared import output


def scan(session: Session):
    if session.args.ports:
        _check_open_ports(session.domain, session.url)


def _check_open_ports(domain: str, url: str, file: Optional[str] = None):
    try:
        output.empty()
        output.norm("Open Ports:")

        ips = basic.get_ips(domain)

        for ip in ips:
            with Spinner():
                res = port_scan.check_open_ports(url, ip, file)

            if len(res) > 0:
                reporter.display_results(res, "\t")
    except Exception as error:
        output.error(f"Error checking for open ports: {str(error)}")



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

from enum import Enum
from typing import NamedTuple

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes


class Severity(str, Enum):
    CRITICAL = "critical"
    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"
    BEST_PRACTICE = "best_practice"
    INFO = "info"


class VulnerabilityInfo(NamedTuple):
    name: str
    severity: Severity
    description: str
    display_all: bool = False
    id: str = "(Invalid)"  # must be the last item

    @classmethod
    def create(
        cls, name: str, severity: Severity, description: str, display_all: bool = False
    ):
        digest = hashes.Hash(hashes.SHAKE128(5), backend=default_backend())
        digest.update(name.encode("utf_8"))
        d = digest.finalize().hex()
        id_val = f"Y{d}"

        return cls.__new__(cls, name, severity, description, display_all, id_val)


class VulnerabilityInfoEnum(VulnerabilityInfo, Enum):
    pass


class Vulnerabilities(VulnerabilityInfoEnum):
    APP_WORDPRESS_VERSION = VulnerabilityInfo.create(
        "App_WordPress_Version", Severity.LOW, ""
    )
    APP_WORDPRESS_OUTDATED = VulnerabilityInfo.create(
        "App_WordPress_Outdated", Severity.MEDIUM, ""
    )
    APP_WORDPRESS_USER_ENUM_API = VulnerabilityInfo.create(
        "App_WordPress_User_Enum_API", Severity.MEDIUM, ""
    )
    APP_WORDPRESS_USER_FOUND = VulnerabilityInfo.create(
        "App_WordPress_User_Found", Severity.LOW, "", True
    )
    APP_WORDPRESS_PATH_DISCLOSURE = VulnerabilityInfo.create(
        "App_WordPress_Path_Disclosure", Severity.LOW, "", True
    )
    APP_JIRA_FOUND = VulnerabilityInfo.create("App_Jira_Found", Severity.INFO, "")
    APP_JIRA_USER_REG_ENABLED = VulnerabilityInfo.create(
        "App_Jira_User_Reg_Enabled", Severity.LOW, ""
    )

    COOKIE_MISSING_SECURE_FLAG = VulnerabilityInfo.create(
        "Cookie_Missing_Secure_Flag", Severity.MEDIUM, "", True
    )
    COOKIE_MISSING_HTTPONLY_FLAG = VulnerabilityInfo.create(
        "Cookie_Missing_HttpOnly_Flag", Severity.LOW, "", True
    )
    COOKIE_MISSING_SAMESITE_FLAG = VulnerabilityInfo.create(
        "Cookie_Missing_SameSite_Flag", Severity.BEST_PRACTICE, "", True
    )
    COOKIE_WITH_SAMESITE_NONE_FLAG = VulnerabilityInfo.create(
        "Cookie_With_SameSite_None_Flag", Severity.BEST_PRACTICE, "", True
    )
    COOKIE_INVALID_SECURE_FLAG = VulnerabilityInfo.create(
        "Cookie_Invalid_Secure_Flag", Severity.MEDIUM, "", True
    )
    COOKIE_INVALID_SAMESITE_NONE_FLAG = VulnerabilityInfo.create(
        "Cookie_Invalid_SameSite_None_Flag", Severity.LOW, "", True
    )
    COOKIE_BIGIP_IP_DISCLOSURE = VulnerabilityInfo.create(
        "Cookie_BigIP_IP_Disclosure", Severity.LOW, "", True
    )

    DNS_CAA_MISSING = VulnerabilityInfo.create("Dns_CAA_Missing", Severity.LOW, "")
    DNS_DNSSEC_NOT_ENABLED = VulnerabilityInfo.create(
        "Dns_DNSSEC_Not_Enabled", Severity.BEST_PRACTICE, ""
    )

    JS_VULNERABLE_VERSION = VulnerabilityInfo.create(
        "Js_Vulnerable_Version", Severity.MEDIUM, "", True
    )
    JS_EXTERNAL_FILE = VulnerabilityInfo.create(
        "Js_External_File", Severity.LOW, "", True
    )

    HTTP_BANNER_GENERIC_APACHE = VulnerabilityInfo.create(
        "Http_Banner_Generic_Apache", Severity.INFO, ""
    )
    HTTP_BANNER_APACHE_VERSION = VulnerabilityInfo.create(
        "Http_Banner_Apache_Version", Severity.LOW, ""
    )
    HTTP_BANNER_GENERIC_NGINX = VulnerabilityInfo.create(
        "Http_Banner_Generic_Nginx", Severity.INFO, ""
    )
    HTTP_BANNER_NGINX_VERSION = VulnerabilityInfo.create(
        "Http_Banner_Nginx_Version", Severity.LOW, ""
    )
    HTTP_BANNER_PYTHON_VERSION = VulnerabilityInfo.create(
        "Http_Banner_Python_Version", Severity.LOW, ""
    )
    HTTP_BANNER_IIS_VERSION = VulnerabilityInfo.create(
        "Http_Banner_IIS_Version", Severity.LOW, ""
    )
    HTTP_BANNER_OPENSSL_VERSION = VulnerabilityInfo.create(
        "Http_Banner_OpenSSL_Version", Severity.LOW, ""
    )
    HTTP_PHP_VERSION_EXPOSED = VulnerabilityInfo.create(
        "Http_PHP_Version_Exposed", Severity.LOW, ""
    )

    HTTP_HEADER_CONTENT_SECURITY_POLICY_MISSING = VulnerabilityInfo.create(
        "Http_Header_Content_Security_Policy_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_CORS_ACAO_UNRESTRICTED = VulnerabilityInfo.create(
        "Http_Header_CORS_ACAO_Unrestricted", Severity.LOW, ""
    )
    HTTP_HEADER_FEATURE_POLICY_MISSING = VulnerabilityInfo.create(
        "Http_Header_Feature_Policy_Missing", Severity.BEST_PRACTICE, ""
    )
    HTTP_HEADER_HSTS_MISSING = VulnerabilityInfo.create(
        "Http_Hsts_Missing", Severity.MEDIUM, ""
    )
    HTTP_HEADER_REFERRER_POLICY_MISSING = VulnerabilityInfo.create(
        "Http_Header_Referrer_Policy_Missing", Severity.BEST_PRACTICE, ""
    )
    HTTP_HEADER_VIA = VulnerabilityInfo.create(
        "Http_Header_Via", Severity.BEST_PRACTICE, ""
    )
    HTTP_HEADER_X_BACKEND_SERVER = VulnerabilityInfo.create(
        "Http_Header_X_Backend_Server", Severity.BEST_PRACTICE, ""
    )
    HTTP_HEADER_X_CONTENT_TYPE_OPTIONS_MISSING = VulnerabilityInfo.create(
        "Http_Header_X_Content_Type_Options_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_X_FRAME_OPTIONS_ALLOW = VulnerabilityInfo.create(
        "Http_Header_X_Frame_Options_Allow", Severity.LOW, ""
    )
    HTTP_HEADER_X_FRAME_OPTIONS_MISSING = VulnerabilityInfo.create(
        "Http_Header_X_Frame_Options_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_X_POWERED_BY = VulnerabilityInfo.create(
        "Http_Header_X_Powered_By", Severity.BEST_PRACTICE, ""
    )
    HTTP_HEADER_X_RUNTIME = VulnerabilityInfo.create(
        "Http_Header_X_Runtime", Severity.BEST_PRACTICE, ""
    )
    HTTP_HEADER_X_XSS_PROTECTION_DISABLED = VulnerabilityInfo.create(
        "Http_Header_X_Xss_Protection_Disabled", Severity.LOW, ""
    )
    HTTP_HEADER_X_XSS_PROTECTION_MISSING = VulnerabilityInfo.create(
        "Http_Header_X_Xss_Protection_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_X_ASPNETMVC_VERSION = VulnerabilityInfo.create(
        "Http_X_AspNetMvc_Version", Severity.LOW, ""
    )
    HTTP_HEADER_X_ASPNET_VERSION = VulnerabilityInfo.create(
        "Http_X_AspNet_Version", Severity.LOW, ""
    )
    HTTP_HEADER_CONTENT_TYPE_NO_CHARSET = VulnerabilityInfo.create(
        "Http_Header_Content_Type_No_Charset", Severity.LOW, ""
    )
    HTTP_HEADER_CONTENT_TYPE_MISSING = VulnerabilityInfo.create(
        "Http_Header_Content_Type_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_CACHE_CONTROL_MISSING = VulnerabilityInfo.create(
        "Http_Header_Cache_Control_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_CACHE_CONTROL_NO_CACHE_MISSING = VulnerabilityInfo.create(
        "Http_Header_Cache_Control_No_Cache_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_CACHE_CONTROL_NO_STORE_MISSING = VulnerabilityInfo.create(
        "Http_Header_Cache_Control_No_Store_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_CACHE_CONTROL_PRIVATE_MISSING = VulnerabilityInfo.create(
        "Http_Header_Cache_Control_Private_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_CACHE_CONTROL_PUBLIC = VulnerabilityInfo.create(
        "Http_Header_Cache_Control_Public", Severity.LOW, ""
    )
    HTTP_HEADER_EXPIRES_MISSING = VulnerabilityInfo.create(
        "Http_Header_Expires_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_PRAGMA_NO_CACHE_MISSING = VulnerabilityInfo.create(
        "Http_Header_Pragma_No_Cache_Missing", Severity.LOW, ""
    )
    HTTP_HEADER_DUPLICATE = VulnerabilityInfo.create(
        "Http_Header_Duplicate", Severity.LOW, ""
    )

    HTTP_ERROR_MESSAGE = VulnerabilityInfo.create(
        "Http_Error_Message", Severity.MEDIUM, "", True
    )
    HTTP_INSECURE_LINK = VulnerabilityInfo.create(
        "Http_Insecure_Link", Severity.LOW, "", True
    )
    HTTP_PROPFIND_ENABLED = VulnerabilityInfo.create(
        "Http_Propfind_Enabled", Severity.LOW, ""
    )
    HTTP_TRACE_ENABLED = VulnerabilityInfo.create(
        "Http_Trace_Enabled", Severity.LOW, ""
    )
    HTTP_OPTIONS_ALLOW = VulnerabilityInfo.create(
        "Http_Option_Allow", Severity.INFO, ""
    )
    HTTP_OPTIONS_PUBLIC = VulnerabilityInfo.create(
        "Http_Option_Public", Severity.INFO, ""
    )
    HTTP_USER_ENUMERATION = VulnerabilityInfo.create(
        "Http_User_Enum_Responses", Severity.MEDIUM, ""
    )
    HTTP_USER_ENUMERATION_TIMING = VulnerabilityInfo.create(
        "Http_User_Enum_Timing", Severity.MEDIUM, ""
    )
    HTTP_BACKUP_FILE = VulnerabilityInfo.create(
        "Http_Backup_File", Severity.BEST_PRACTICE, "", True
    )
    HTTP_DS_STORE_FILE = VulnerabilityInfo.create(
        "Http_DS_Store_File", Severity.INFO, "", True
    )

    TLS_CBC_CIPHER_SUITE = VulnerabilityInfo.create(
        "Tls_CBC_Cipher_Suite", Severity.BEST_PRACTICE, ""
    )

    TLS_CERT_BAD_COMMON_NAME = VulnerabilityInfo.create(
        "Tls_Cert_Bad_Common_Name", Severity.HIGH, ""
    )
    TLS_CERT_BLACKLISTED = VulnerabilityInfo.create(
        "Tls_Cert_Blacklisted", Severity.HIGH, ""
    )
    TLS_CERT_EXPIRED = VulnerabilityInfo.create("Tls_Cert_Expired", Severity.HIGH, "")
    TLS_CERT_HOSTNAME_MISMATCH = VulnerabilityInfo.create(
        "Tls_Cert_Hostname_Mismatch", Severity.HIGH, ""
    )
    TLS_CERT_INSECURE_KEY = VulnerabilityInfo.create(
        "Tls_Cert_Insecure_Key", Severity.HIGH, ""
    )
    TLS_CERT_INSECURE_SIGNATURE = VulnerabilityInfo.create(
        "Tls_Cert_Insecure_Signature", Severity.HIGH, ""
    )
    TLS_CERT_NOT_YET_VALID = VulnerabilityInfo.create(
        "Tls_Cert_Not_Yet_Valid", Severity.HIGH, ""
    )
    TLS_CERT_NO_TRUST = VulnerabilityInfo.create("Tls_Cert_No_Trust", Severity.HIGH, "")
    TLS_CERT_REVOKED = VulnerabilityInfo.create("Tls_Cert_Revoked", Severity.HIGH, "")
    TLS_CERT_SELF_SIGNED = VulnerabilityInfo.create(
        "Tls_Cert_Self_Signed", Severity.HIGH, ""
    )

    TLS_COMPRESSION_ENABLED = VulnerabilityInfo.create(
        "Tls_Compression_Enabled", Severity.HIGH, "", True
    )
    TLS_DH_KNOWN_PRIMES_STRONG = VulnerabilityInfo.create(
        "Tls_DH_Known_Primes_Strong", Severity.MEDIUM, "", True
    )
    TLS_DH_KNOWN_PRIMES_WEAK = VulnerabilityInfo.create(
        "Tls_DH_Known_Primes_Weak", Severity.HIGH, "", True
    )
    TLS_DH_PARAM_REUSE = VulnerabilityInfo.create(
        "Tls_DH_Param_Reuse", Severity.LOW, "", True
    )
    TLS_DROWN = VulnerabilityInfo.create("Tls_Drown", Severity.MEDIUM, "", True)
    TLS_ECDH_PARAM_REUSE = VulnerabilityInfo.create(
        "Tls_ECDH_Param_Reuse", Severity.LOW, "", True
    )
    TLS_FALLBACK_SCSV_MISSING = VulnerabilityInfo.create(
        "Tls_Fallback_SCSV_Missing", Severity.LOW, "", True
    )
    TLS_FREAK = VulnerabilityInfo.create("Tls_Freak", Severity.HIGH, "", True)
    TLS_GOLDENDOODLE = VulnerabilityInfo.create(
        "Tls_Goldendoodle", Severity.HIGH, "", True
    )
    TLS_GOLDENDOODLE_NE = VulnerabilityInfo.create(
        "Tls_Goldendoodle_NE", Severity.MEDIUM, "", True
    )
    TLS_HEARTBEAT_ENABLED = VulnerabilityInfo.create(
        "Tls_Heartbeat_Enabled", Severity.BEST_PRACTICE, "", True
    )
    TLS_HEARTBLEED = VulnerabilityInfo.create(
        "Tls_Heartbleed", Severity.CRITICAL, "", True
    )
    TLS_INSECURE_CIPHER_SUITE = VulnerabilityInfo(
        "Tls_Insecure_Cipher_Suite", Severity.MEDIUM, "", True
    )
    TLS_INSECURE_RENEG = VulnerabilityInfo.create(
        "Tls_Insecure_Reneg", Severity.HIGH, "", True
    )
    TLS_LEGACY_SSL_ENABLED = VulnerabilityInfo.create(
        "Tls_Legacy_SSL_Enabled", Severity.HIGH, "", True
    )
    TLS_LEGACY_SSL_POODLE = VulnerabilityInfo.create(
        "Tls_Legacy_SSL_Poodle", Severity.HIGH, "", True
    )
    TLS_LIMITED_FORWARD_SECRECY = VulnerabilityInfo.create(
        "Tls_Limited_Forward_Secrecy", Severity.LOW, "", True
    )
    TLS_LOGJAM = VulnerabilityInfo.create("Tls_Logjam", Severity.HIGH, "", True)
    TLS_NO_AEAD_SUPPORT = VulnerabilityInfo.create(
        "Tls_No_AEAD_Support", Severity.BEST_PRACTICE, "", True
    )
    TLS_OCSP_STAPLE_MISSING = VulnerabilityInfo.create(
        "Tls_OCSP_Staple_Missing", Severity.LOW, "", True
    )

    TLS_OPENSSL_CVE_2014_0224 = VulnerabilityInfo.create(
        "Tls_OpenSSL_CVE_2014_0224", Severity.HIGH, "", True
    )
    TLS_OPENSSL_CVE_2014_0224_NE = VulnerabilityInfo.create(
        "Tls_OpenSSL_CVE_2014_0224_NE", Severity.MEDIUM, "", True
    )
    TLS_OPENSSL_CVE_2016_2107 = VulnerabilityInfo.create(
        "Tls_OpenSSL_CVE_2016_2107", Severity.HIGH, "", True
    )
    TLS_OPENSSL_CVE_2019_1559 = VulnerabilityInfo.create(
        "Tls_OpenSSL_CVE_2019_1559", Severity.HIGH, "", True
    )
    TLS_OPENSSL_CVE_2019_1559_NE = VulnerabilityInfo.create(
        "Tls_OpenSSL_CVE_2019_1559_NE", Severity.MEDIUM, "", True
    )

    TLS_POODLE = VulnerabilityInfo.create("Tls_Poodle", Severity.HIGH, "", True)
    TLS_ROBOT_ORACLE_STRONG = VulnerabilityInfo.create(
        "Tls_Robot_Oracle_Strong", Severity.MEDIUM, "", True
    )
    TLS_ROBOT_ORACLE_WEAK = VulnerabilityInfo.create(
        "Tls_Robot_Oracle_Weak", Severity.LOW, "", True
    )
    TLS_SESSION_RESP_ENABLED = VulnerabilityInfo.create(
        "Tls_Session_Resp_Enabled", Severity.BEST_PRACTICE, "", True
    )
    TLS_SLEEPING_POODLE = VulnerabilityInfo.create(
        "Tls_Sleeping_Poodle", Severity.HIGH, "", True
    )
    TLS_SLEEPING_POODLE_NE = VulnerabilityInfo.create(
        "Tls_Sleeping_Poodle_NE", Severity.MEDIUM, "", True
    )
    TLS_SWEET32 = VulnerabilityInfo.create("Tls_SWEET32", Severity.HIGH, "", True)
    TLS_SYMANTEC_ROOT = VulnerabilityInfo.create(
        "Tls_Symantec_Root", Severity.HIGH, "", True
    )
    TLS_TICKETBLEED = VulnerabilityInfo.create(
        "Tls_Ticketbleed", Severity.HIGH, "", True
    )

    TLS_VERSION_1_0_ENABLED = VulnerabilityInfo.create(
        "Tls_Version_1_0_Enabled", Severity.LOW, "", True
    )
    TLS_VERSION_1_3_EARLY_DATA_ENABLED = VulnerabilityInfo.create(
        "Tls_Version_1_3_Early_Data_Enabled", Severity.BEST_PRACTICE, "", True
    )
    TLS_VERSION_1_3_NOT_ENABLED = VulnerabilityInfo.create(
        "Tls_Version_1_3_Not_Enabled", Severity.BEST_PRACTICE, "", True
    )

    TLS_ZOMBIE_POODLE = VulnerabilityInfo.create(
        "Tls_Zombie_Poodle", Severity.HIGH, "", True
    )
    TLS_ZOMBIE_POODLE_NE = VulnerabilityInfo.create(
        "Tls_Zombie_Poodle_NE", Severity.MEDIUM, "", True
    )

    SERVER_APACHE_OUTDATED = VulnerabilityInfo.create(
        "Server_Apache_Outdated", Severity.MEDIUM, ""
    )
    SERVER_APACHE_STATUS = VulnerabilityInfo.create(
        "Server_Apache_Status", Severity.MEDIUM, ""
    )
    SERVER_APACHE_INFO = VulnerabilityInfo.create(
        "Server_Apache_Info", Severity.MEDIUM, ""
    )
    SERVER_TOMCAT_VERSION = VulnerabilityInfo.create(
        "Server_Tomcat_Version", Severity.MEDIUM, ""
    )
    SERVER_TOMCAT_OUTDATED = VulnerabilityInfo.create(
        "Server_Tomcat_Outdated", Severity.MEDIUM, ""
    )
    SERVER_TOMCAT_MANAGER_EXPOSED = VulnerabilityInfo.create(
        "Server_Tomcat_Manager_Exposed", Severity.HIGH, ""
    )
    SERVER_TOMCAT_HOST_MANAGER_EXPOSED = VulnerabilityInfo.create(
        "Server_Tomcat_Host_Manager_Exposed", Severity.HIGH, ""
    )
    SERVER_TOMCAT_MANAGER_WEAK_PASSWORD = VulnerabilityInfo.create(
        "Server_Tomcat_Manager_Weak_Password", Severity.CRITICAL, "", True
    )
    SERVER_TOMCAT_CVE_2017_12615 = VulnerabilityInfo.create(
        "Server_Tomcat_CVE_2017_12615", Severity.CRITICAL, ""
    )
    SERVER_TOMCAT_CVE_2019_0232 = VulnerabilityInfo.create(
        "Server_Tomcat_CVE_2019_0232", Severity.CRITICAL, ""
    )
    SERVER_TOMCAT_STRUTS_SAMPLE = VulnerabilityInfo.create(
        "Server_Tomcat_Struts_Sample", Severity.LOW, "", True
    )
    SERVER_NGINX_OUTDATED = VulnerabilityInfo.create(
        "Server_Nginx_Outdated", Severity.MEDIUM, ""
    )
    SERVER_NGINX_STATUS_EXPOSED = VulnerabilityInfo.create(
        "Server_Nginx_Status_Exposed", Severity.LOW, ""
    )
    SERVER_PHP_OUTDATED = VulnerabilityInfo.create(
        "Server_PHP_Outdated", Severity.MEDIUM, ""
    )
    SERVER_PHP_PHPINFO = VulnerabilityInfo.create(
        "Server_PHP_PHPInfo", Severity.LOW, ""
    )
    SERVER_PHP_CVE_2019_11043 = VulnerabilityInfo.create(
        "Server_PHP_CVE_2019_11043", Severity.CRITICAL, ""
    )
    SERVER_IIS_OUTDATED = VulnerabilityInfo.create(
        "Server_IIS_Outdated", Severity.MEDIUM, ""
    )
    SERVER_ASPNETMVC_OUTDATED = VulnerabilityInfo.create(
        "Server_AspNetMvc_Outdated", Severity.MEDIUM, ""
    )
    SERVER_ASPNET_OUTDATED = VulnerabilityInfo.create(
        "Server_AspNet_Outdated", Severity.MEDIUM, ""
    )
    SERVER_ASPNET_DEBUG_ENABLED = VulnerabilityInfo.create(
        "Server_AspNet_Debug_Enabled", Severity.HIGH, ""
    )
    SERVER_ASPNET_HANDLER_ENUM = VulnerabilityInfo.create(
        "Server_AspNet_Handler_Enum", Severity.LOW, "", True
    )
    SERVER_RAILS_CVE_2019_5418 = VulnerabilityInfo.create(
        "Server_Rails_CVE_2019_5418", Severity.CRITICAL, ""
    )
    SERVER_INVALID_404_FILE = VulnerabilityInfo.create(
        "Server_Invalid_404_File", Severity.INFO, ""
    )
    SERVER_INVALID_404_PATH = VulnerabilityInfo.create(
        "Server_Invalid_404_Path", Severity.INFO, ""
    )
    SERVER_SPECIAL_FILE_EXPOSED = VulnerabilityInfo.create(
        "Server_Special_File_Exposed", Severity.INFO, "", True
    )
    SERVER_INT_IP_EXP_HTTP10 = VulnerabilityInfo.create(
        "Server_Int_IP_Exp_Http10", Severity.LOW, ""
    )

    NETWORK_OPEN_PORT = VulnerabilityInfo.create(
        "Network_Open_Port", Severity.INFO, "", True
    )

    WAF_CLOUDFLARE = VulnerabilityInfo.create("Waf_Cloudflare", Severity.INFO, "")
    WAF_INCAPSULA = VulnerabilityInfo.create("Waf_Incapsula", Severity.INFO, "")



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

from datetime import datetime


class ExecutionTimer:
    def __enter__(self):
        self.begin = datetime.now()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = datetime.now()
        self.execution_time = self.end - self.begin

    def to_ms(self) -> int:
        """
        Returns the execution time, in milliseconds
        :return:
        """
        return int(self.execution_time.total_seconds() * 1000)



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

import argparse
import sys
from typing import List

from yawast._version import get_version
from yawast.commands import scan, dns, ssl
from yawast.reporting import reporter
from yawast.scanner.session import Session
from yawast.shared import utils


def build_parser():
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "--debug", action="store_true", help="Displays debug output (very noisy)"
    )
    parent_parser.add_argument(
        "--nocolors", action="store_true", help="Disables the use of colors in output"
    )
    parent_parser.add_argument(
        "--nowrap",
        action="store_true",
        help="Disables the use of line wrapping in output",
    )

    parser = argparse.ArgumentParser(prog="yawast")
    parser.description = "YAWAST ...where a pentest starts. A tool for web-based application security testing."
    parser.epilog = "For more information, see https://yawast.org"
    parser.add_argument(
        "-v", "--version", action="version", version=f"{parser.prog} v{get_version()}"
    )

    subparsers = parser.add_subparsers()
    subparsers.required = True
    subparsers.dest = "command"

    # create the parser for the "scan" command
    parser_scan = subparsers.add_parser(
        "scan", help="Scans the provided URL(s)", parents=[parent_parser]
    )
    parser_scan.add_argument("--nossl", action="store_true", help="Disables SSL checks")
    parser_scan.add_argument(
        "--internalssl", action="store_true", help="Disable SSL Labs integration"
    )
    parser_scan.add_argument(
        "--tdessessioncount",
        action="store_true",
        help="Counts the number of messages that can be sent in a single session (SWEET32)",
    )
    parser_scan.add_argument(
        "--dir", action="store_true", help="Enables directory search"
    )
    parser_scan.add_argument(
        "--dirrecursive",
        action="store_true",
        help="Recursive directory search (only with --dir)",
    )
    parser_scan.add_argument(
        "--dirlistredir",
        action="store_true",
        help="Show 301 redirects (only with --dir)",
    )
    parser_scan.add_argument(
        "--files",
        action="store_true",
        help="Performs a search for a large list of common files",
    )
    parser_scan.add_argument(
        "--srv", action="store_true", help="Scan for known SRV DNS Records"
    )
    parser_scan.add_argument(
        "--subdomains", action="store_true", help="Search for Common Subdomains"
    )
    parser_scan.add_argument("--nodns", action="store_true", help="Disable DNS checks")
    parser_scan.add_argument(
        "--ports", action="store_true", help="Scan common TCP ports"
    )
    parser_scan.add_argument(
        "--proxy", type=str, help="HTTP Proxy Server (such as Burp Suite)"
    )
    parser_scan.add_argument("--cookie", type=str, help="Session cookie")
    parser_scan.add_argument(
        "--header",
        type=str,
        help="HTTP header (such as Authorization) sent with each request ('name=value')",
    )
    parser_scan.add_argument("--output", type=str, help="Output JSON file")
    parser_scan.add_argument(
        "--user",
        type=str,
        help="Valid username for the application (will prompt if not provided)",
    )
    parser_scan.add_argument(
        "--pass_reset_page",
        type=str,
        help="Password reset page URL (will prompt if not provided)",
    )
    parser_scan.add_argument(
        "--php_page",
        type=str,
        help="Relative path to PHP script (for additional tests)",
    )
    parser_scan.set_defaults(func=command_scan)

    # create the parser for the "dns" command
    parser_dns = subparsers.add_parser(
        "dns", help="Scans DNS for the provided URL(s)", parents=[parent_parser]
    )
    parser_dns.add_argument(
        "--srv", action="store_true", help="Scan for known SRV DNS Records"
    )
    parser_dns.add_argument(
        "--subdomains", action="store_true", help="Search for Common Subdomains"
    )
    parser_dns.add_argument("--output", type=str, help="Output JSON file")
    parser_dns.set_defaults(func=command_dns)

    # create the parser for the "ssl" command
    parser_ssl = subparsers.add_parser(
        "ssl", help="Scans TLS/SSL for the provided URL(s)", parents=[parent_parser]
    )
    parser_ssl.add_argument(
        "--internalssl", action="store_true", help="Disable SSL Labs integration"
    )
    parser_ssl.add_argument(
        "--tdessessioncount",
        action="store_true",
        help="Counts the number of messages that can be sent in a single session (SWEET32)",
    )
    parser_ssl.add_argument("--nodns", action="store_true", help="Disable DNS checks")
    parser_ssl.add_argument("--output", type=str, help="Output JSON file")
    parser_ssl.set_defaults(func=command_ssl)

    # create the parser for the "version" command
    parser_version = subparsers.add_parser(
        "version",
        help="Displays information about YAWAST and the current environment",
        parents=[parent_parser],
    )
    parser_version.add_argument("--output", type=str, help="Output JSON file")
    parser_version.set_defaults(func=command_version)

    return parser


def process_urls(urls) -> List[str]:
    ret = []

    # now we need to make we have at least one arg that could be a URL.
    if len(urls) == 0:
        utils.exit_message("YAWAST Error: You must specify at least one URL.")

    # Next, we need to make sure we have something that looks like URLs.
    for val in enumerate(urls):
        if not str(val[1]).startswith("-"):
            if not utils.is_url(val[1]):
                utils.exit_message("YAWAST Error: Invalid URL Specified: '%s" % val[1])
            else:
                ret.append(val[1])
        else:
            print(
                "YAWAST Error: Invalid parameter: '%s' - Ignored." % val[1],
                file=sys.stderr,
            )

    return ret


def command_scan(args, urls):
    for val in enumerate(urls):
        url = utils.extract_url(val[1])

        reporter.setup(utils.get_domain(url))

        session = Session(args, url)

        scan.start(session)


def command_dns(args, urls):
    for val in enumerate(urls):
        url = utils.extract_url(val[1])

        reporter.setup(utils.get_domain(url))

        session = Session(args, url)

        dns.start(session)


def command_ssl(args, urls):
    for val in enumerate(urls):
        url = utils.extract_url(val[1])

        reporter.setup(utils.get_domain(url))

        session = Session(args, url)

        ssl.start(session)


def command_version(args, urls):
    pass



#  Copyright (c) 2013 - 2019 Adam Caudill and Contributors.
#  This file is part of YAWAST which is released under the MIT license.
#  See the LICENSE file or go to https://yawast.org/license/ for full license details.

import base64
import re
import secrets
from typing import List, cast, Union, Optional
from urllib.parse import urljoin

from packaging import version
from requests import Response

from yawast.reporting.enums import Vulnerabilities
from yawast.scanner.plugins.http import version_checker, response_scanner
from yawast.scanner.plugins.result import Result
from yawast.shared import network, output


def check_all(url: str, links: List[str]) -> List[Result]:
    results: List[Result] = []

    results += check_version(url)
    results += check_manager(url)
    results += check_cve_2017_12615(url)
    results += check_struts_sample(url)
    results += check_cve_2019_0232(links)

    return results


def get_version(url: str, res: Response, method: Optional[str] = None) -> List[Result]:
    """Check a server response to see if it contains a Tomcat version.

    :param method:
    :param url:
    :param res:
    :return:
    """

    results: List[Result] = []

    try:
        body = res.text
        ver = _get_version_from_body(body, res.status_code)
        if ver is not None:
            msg = f"Apache Tomcat version exposed: {ver}"
            if method is not None:
                msg += f" (Via {method})"

            results.append(
                Result(
                    msg,
                    Vulnerabilities.SERVER_TOMCAT_VERSION,
                    url,
                    [
                        ver,
                        network.http_build_raw_request(res.request),
                        network.http_build_raw_response(res),
                    ],
                )
            )

            results += _check_version_outdated(ver, url, body)
    except Exception:
        output.debug_exception()

    return results


def check_version(url: str) -> List[Result]:
    """Search for the version of Tomcat used, by using various methods.
    Methods include sending a POST request to a page that likely doesn't
    support POST, sending an invalid HTTP verb, and triggering a 404.

    :param url:
    :return:
    """

    results: List[Result] = []

    results += _check_version_404(url)
    results += _check_version_verb(url)
    results += _check_version_post(url)

    return results


def check_manager(url: str) -> List[Result]:
    results = []

    try:
        for p in [urljoin(url, "manager/"), urljoin(url, "host-manager/")]:
            # check for both Tomcat 6, and 7+
            for path in [urljoin(p, "html/"), p]:

                res = network.http_get(path, False)

                body = res.text

                if "<tt>conf/tomcat-users.xml</tt>" in body:
                    # we have a finding
                    vuln = Vulnerabilities.SERVER_TOMCAT_MANAGER_EXPOSED
                    if "host-manager" in path:
                        vuln = Vulnerabilities.SERVER_TOMCAT_HOST_MANAGER_EXPOSED

                    results.append(
                        Result(
                            f"Apache Tomcat Manager found: {path}",
                            vuln,
                            path,
                            [
                                network.http_build_raw_request(res.request),
                                network.http_build_raw_response(res),
                            ],
                        )
                    )

                    # check to see if we can get in with a default password
                    results += check_manager_password(url)
                else:
                    # if we didn't get a hit, go ahead and scan it to see if there's
                    #  anything else that we should be picking up.
                    results += response_scanner.check_response(path, res)
    except Exception:
        output.debug_exception()

    return results


def check_manager_password(url: str) -> List[Result]:
    results = []

    try:
        creds = [
            b"tomcat:tomcat",
            b"tomcat:password",
            b"tomcat:",
            b"admin:admin",
            b"admin:password",
            b"admin:",
        ]

        for cred in creds:
            ce = base64.b64encode(cred)

            res = network.http_get(url, False, {"Authorization": ce})
            body = res.text

            if (
                '<font size="+2">Tomcat Web Application Manager</font>' in body
                or '<font size="+2">Tomcat Virtual Host Manager</font>' in body
            ):
                # we got in
                results.append(
                    Result(
                        f"Apache Tomcat Weak Manager Password: '{cred}' - {url}",
                        Vulnerabilities.SERVER_TOMCAT_MANAGER_WEAK_PASSWORD,
                        url,
                        [
                            network.http_build_raw_request(res.request),
                            network.http_build_raw_response(res),
                        ],
                    )
                )
            else:
                # if we didn't get a hit, go ahead and scan it to see if there's
                #  anything else that we should be picking up.
                results += response_scanner.check_response(url, res)
    except Exception:
        output.debug_exception()

    return results


def check_cve_2017_12615(url: str) -> List[Result]:
    results = []

    try:
        file_name = secrets.token_hex(12)
        check_value = secrets.token_hex(12)

        target = urljoin(url, f"{file_name}.jsp/")
        res_put = network.http_put(target, f"<% out.println({check_value});%>", False)

        if res_put.status_code < 300:
            # code should be 2xx for this to work
            # now we need to check to see if it worked
            created_file = urljoin(url, f"{file_name}.jsp")

            res_get = network.http_get(created_file, False)

            if check_value in res_get.text:
                # we have RCE
                results.append(
                    Result(
                        f"Apache Tomcat PUT RCE (CVE-2017-12615): {created_file}",
                        Vulnerabilities.SERVER_TOMCAT_CVE_2017_12615,
                        url,
                        [
                            network.http_build_raw_request(res_put.request),
                            network.http_build_raw_response(res_put),
                            network.http_build_raw_request(res_get.request),
                            network.http_build_raw_response(res_get),
                        ],
                    )
                )
            else:
                results += response_scanner.check_response(created_file, res_get)
        else:
            # if we didn't get a hit, go ahead and scan it to see if there's
            #  anything else that we should be picking up.
            results += response_scanner.check_response(target, res_put)
    except Exception:
        output.debug_exception()

    return results


def check_cve_2019_0232(links: List[str]) -> List[Result]:
    results: List[Result] = []

    try:
        targets: List[str] = []

        for link in links:
            if "cgi-bin" in link:
                if "?" in link:
                    targets.append(f"{link}&dir")
                else:
                    targets.append(f"{link}?dir")

        for target in targets:
            res = network.http_get(target, False)
            body = res.text

            if "<DIR>" in body:
                # we have a hit
                results.append(
                    Result(
                        f"Apache Tomcat RCE (CVE-2019-0232): {target}",
                        Vulnerabilities.SERVER_TOMCAT_CVE_2019_0232,
                        target,
                        [
                            network.http_build_raw_request(res.request),
                            network.http_build_raw_response(res),
                        ],
                    )
                )

            results += response_scanner.check_response(target, res)
    except Exception:
        output.debug_exception()

    return results


def check_struts_sample(url: str) -> List[Result]:
    results: List[Result] = []

    try:
        # make sure we have real 404s
        file_good, _, _, _ = network.check_404_response(url)
        if not file_good:
            return results

        search = [
            "Struts2XMLHelloWorld/User/home.action",
            "struts2-showcase/showcase.action",
            "struts2-showcase/titles/index.action",
            "struts2-bootstrap-showcase/",
            "struts2-showcase/index.action",
            "struts2-bootstrap-showcase/index.action",
            "struts2-rest-showcase/",
        ]

        for path in search:
            target = urljoin(url, path)

            res = network.http_get(target, False)

            # check for other issues
            results += response_scanner.check_response(target, res)

            if res.status_code == 200:
                results.append(
                    Result(
                        f"Struts Sample Found: {target}",
                        Vulnerabilities.SERVER_TOMCAT_STRUTS_SAMPLE,
                        target,
                        [
                            network.http_build_raw_request(res.request),
                            network.http_build_raw_response(res),
                        ],
                    )
                )
    except Exception:
        output.debug_exception()

    return results


def _check_version_404(url: str) -> List[Result]:
    results: List[Result] = []

    try:
        rnd = secrets.token_hex(12)

        target = urljoin(url, f"{rnd}.jsp")

        res = network.http_get(target, False)

        if res.status_code > 400:
            results += get_version(target, res, "404 Error Message")
    except Exception:
        output.debug_exception()

    return results


def _check_version_verb(url: str) -> List[Result]:
    results: List[Result] = []

    try:
        res = network.http_custom("XYZ", url)

        if res.status_code > 400:
            results += get_version(url, res, "Invalid HTTP Verb")
    except Exception:
        output.debug_exception()

    return results


def _check_version_post(url: str) -> List[Result]:
    results: List[Result] = []

    try:
        res = network.http_custom("POST", url)

        if res.status_code > 400:
            results += get_version(url, res, "POST to root")
    except Exception:
        output.debug_exception()

    return results


def _get_version_from_body(body, status_code) -> Union[str, None]:
    if "Apache Tomcat" in body and status_code >= 400:
        # we only care if:
        #  'Apache Tomcat' is in the body
        #  We have a status code that indicates an error
        ver = re.search(r"Apache Tomcat/\d*.\d*.\d*\b", body)

        if ver:
            # split it, and only return the number itself
            return ver.group(0).split("/")[1]

    return None


def _check_version_outdated(ver: str, url: str, body: str) -> List[Result]:
    results: List[Result] = []

    # parse the version, and get the latest version - see if the server is up to date
    ver = cast(version.Version, version.parse(ver))
    curr_version = version_checker.get_latest_version("apache_tomcat", ver)

    if curr_version is not None and curr_version > ver:
        results.append(
            Result(
                f"Apache Tomcat Outdated: {ver} - Current: {curr_version}",
                Vulnerabilities.SERVER_TOMCAT_OUTDATED,
                url,
                [ver, curr_version, body],
            )
        )

    return results
